#!/usr/bin/python3
"""
// © University of Southampton IT Innovation Centre, 2018
//
// Copyright in this software belongs to University of Southampton
// IT Innovation Centre of Gamma House, Enterprise Road,
// Chilworth Science Park, Southampton, SO16 7NS, UK.
//
// This software may not be used, sold, licensed, transferred, copied
// or reproduced in whole or in part in any manner or form or in or
// on any media by any person other than in accordance with the terms
// of the Licence Agreement supplied with the software, or otherwise
// without the prior written consent of the copyright owners.
//
// This software is distributed WITHOUT ANY WARRANTY, without even the
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE, except where stated in the Licence Agreement supplied with
// the software.
//
//      Created By :            Nikolay Stanchev
//      Created Date :          16-08-2018
//      Created for Project :   FLAME
"""


# Python standard libs
from re import compile, IGNORECASE

# PIP installed libs
from schema import Schema, And, Or, Optional


"""
This module defines the schema objects for the TOSCA Alert Specification:

        * flame_clmc_alerts_definitions.yaml must be the only import
        * metadata section must be present (with key-value pairs for sfc and sfci)
        * policies section must be present (under the topology_template node)
        * each policy must be associated with a triggers node (containing at least 1 trigger)
        * each policy is of type eu.ict-flame.policies.StateChange or eu.ict-flame.policies.Alert
        * each trigger must specify event_type, metric, condition, and at least one handler in action/implementation
        * the condition section must specify threshold, granularity, aggregation_method, comparison_operator
"""


# Influx QL functions defined in the documentation https://docs.influxdata.com/influxdb/v1.6/query_language/functions/
INFLUX_QL_FUNCTIONS = (
    "count", "mean", "median", "mode", "sum", "first", "last", "max", "min"
)

# Kapacitor Tick Script template IDs
TICK_SCRIPT_TEMPLATES = ("threshold", "relative", "deadman")

# Allowed comparison operators and their logical values
COMPARISON_OPERATORS = {"lt": "<", "gt": ">", "lte": "<=", "gte": ">=", "eq": "=", "neq": "<>"}

# Regular expression for validating http handlers
URL_REGEX = compile(
    r'^https?://'  # http:// or https://
    r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|'  # domain, e.g. example.domain.com
    r'localhost|'  # or localhost...
    r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})'  # or IP address (IPv4 format)
    r'(?::\d{2,5})?'  # optional port number
    r'(?:[/?#][^\s]*)?$',  # URL path or query parameters
    IGNORECASE)


# Global tags allowed to be used for filtering in the trigger condition
CLMC_INFORMATION_MODEL_GLOBAL_TAGS = ("sfc", "sfci", "sf_package", "sf", "sf_endpoint", "host", "location")


ALERTS_SPECIFICATION_SCHEMA = Schema({
    "tosca_definitions_version": And(str, lambda v: v == "tosca_simple_profile_for_nfv_1_0_0"),
    Optional("description"): str,
    "imports": And([lambda s: s.endswith("flame_clmc_alerts_definitions.yaml")], lambda l: len(l) == 1),
    "metadata": {
        "sfc": str,
        "sfci": str
    },
    "topology_template": {
        "policies": [
            {
                str: {
                    "type": Or("eu.ict-flame.policies.StateChange", "eu.ict-flame.policies.Alert"),
                    "triggers": And({
                        str: {
                            Optional("description"): str,
                            "event_type": And(str, lambda s: s in TICK_SCRIPT_TEMPLATES),
                            "metric": And(str, lambda s: len(s.split('.', 1)) == 2),
                            "condition": {
                                "threshold": Or(int, float),
                                "granularity": int,
                                "aggregation_method": And(str, lambda s: s in INFLUX_QL_FUNCTIONS),
                                Optional("resource_type"): {
                                    And(str, lambda s: s in CLMC_INFORMATION_MODEL_GLOBAL_TAGS): str
                                },
                                "comparison_operator": And(str, lambda s: s in COMPARISON_OPERATORS)
                            },
                            "action": {
                                "implementation":
                                    [
                                        And(str, lambda s: URL_REGEX.match(s) is not None)
                                    ]
                            }
                        }
                    }, lambda l: len(l) > 0)
                }
            }
        ]
    }
})


def fill_http_post_handler_vars(handler_id, handler_url):
    """
    Creates a dictionary object ready to be posted to kapacitor to create an alert handler.

    :param handler_id: handler identifier
    :param handler_url: url to post alerts to

    :return: a dictionary object ready to be posted to kapacitor to create an alert handler.
    """

    return {
        "id": handler_id,
        "kind": "post",
        "options": {
            "url": handler_url
        }
    }


def fill_threshold_template_vars(db, measurement, field, influx_function, critical_value, comparison_operator, alert_period, topic_id, where_clause=None):
    """
    Creates a dictionary object ready to be posted to kapacitor to create a task from template.

    :param db: db name
    :param measurement: measurement name
    :param field: field name
    :param influx_function: influx function to use for querying
    :param critical_value: critical value to compare with
    :param comparison_operator: type of comparison
    :param alert_period: alert period to query influx
    :param topic_id: topic identifier
    :param where_clause: optional argument for filtering the influx query by tag values

    :return: a dictionary object ready to be posted to kapacitor to create a task from template.
    """

    comparison_lambda = "\"real_value\" {0} {1}".format(comparison_operator, critical_value)  # build up lambda string, e.g. "real_value" >= 10

    tempalte_vars = {
        "db": {
            "type": "string",
            "value": db
        },
        "measurement": {
            "type": "string",
            "value": measurement
        },
        "field": {
            "type": "string",
            "value": field
        },
        "influxFunction": {
            "type": "string",
            "value": influx_function
        },
        "comparisonLambda": {
            "type": "lambda",
            "value": comparison_lambda
        },
        "alertPeriod": {
            "type": "duration",
            "value": alert_period
        },
        "topicID": {
            "type": "string",
            "value": topic_id
        }
    }

    if where_clause is not None:
        tempalte_vars["whereClause"] = {
            "type": "string",
            "value": where_clause
        }

    return tempalte_vars
